from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.dispatch import receiver
from signals import m2m_history_changed


class ManyToManyHistoryVersion(models.Model):
    class Meta:
        unique_together = ('content_type', 'object_id', 'field_name', 'time')

    content_type = models.ForeignKey(ContentType, related_name='m2m_history_versions', db_index=True)
    object_id = models.BigIntegerField(db_index=True)
    object = generic.GenericForeignKey('content_type', 'object_id')

    field_name = models.CharField(max_length=50, db_index=True)
    time = models.DateTimeField(db_index=True)

    count = models.PositiveIntegerField(default=0)
    added_count = models.PositiveIntegerField(default=0)
    removed_count = models.PositiveIntegerField(default=0)

    @property
    def items(self):
        return getattr(self.object, self.field_name).were_at(self.time)

    @property
    def added(self):
        return getattr(self.object, self.field_name).added_at(self.time)

    @property
    def removed(self):
        return getattr(self.object, self.field_name).removed_at(self.time)

@receiver(m2m_history_changed)
def save_m2m_history_cache(sender, action, instance, reverse, pk_set, field_name, time, **kwargs):
    need_to_cache = not reverse and instance._meta.get_field(field_name).cache
    if need_to_cache and action in ['post_add', 'post_remove', 'post_clear'] and len(pk_set):
        cache = ManyToManyHistoryVersion.objects.get_or_create(content_type=ContentType.objects.get_for_model(instance), object_id=instance.pk, field_name=field_name, time=time)[0]
        cache.count = getattr(instance, field_name).get_query_set(only_pk=True).count()
        if action in ['post_add']:
            cache.added_count = len(pk_set)
        elif action in ['post_remove', 'post_clear']:
            cache.removed_count = len(pk_set)
        cache.save()


# -*- coding: utf-8 -*-
from django.test import TestCase
from django.db import models
from fields import ManyToManyHistoryField
from datetime import datetime
from models import ManyToManyHistoryVersion
import time

'''
Example model from Django docs: https://docs.djangoproject.com/en/dev/topics/db/examples/many_to_many/
'''

class Publication(models.Model):
    title = models.CharField(max_length=30)


class Article(models.Model):
    headline = models.CharField(max_length=100)
    publications = ManyToManyHistoryField(Publication, cache=True)
    publications_no_cache = ManyToManyHistoryField(Publication, related_name='articles_no_cache')


'''
Test class
'''

class ManyToManyHistoryTest(TestCase):

    def test_m2m_fields_and_methods(self):
        self.assertItemsEqual([field.name for field in Article._meta.get_field('publications').rel.through._meta.local_fields], [u'id', 'time_from', 'time_to', 'article', 'publication'])

    def test_m2m_history_features(self):

        p1 = Publication.objects.create(title='Pub1')
        p2 = Publication.objects.create(title='Pub2')
        p3 = Publication.objects.create(title='Pub3')

        article = Article.objects.create(headline='Article1')
        state_time1 = datetime.now()
        # we need to use sleep here to pass travis mysql tests, becouse Django + mysql doesn't support storing microseconds
        # and as result our state_timeX will be equal
        time.sleep(1)

        article.publications = [p1, p2]
        state_time2 = article.publications.last_update_time()
        self.assertItemsEqual(article.publications.all(), [p1, p2])
        self.assertEqual(article.publications.count(), 2)
        self.assertEqual(article.publications.through.objects.count(), 2)
        time.sleep(1)

        article.publications = [p3]
        state_time3 = article.publications.last_update_time()
        self.assertItemsEqual(article.publications.all(), [p3])
        self.assertEqual(article.publications.count(), 1)
        self.assertEqual(article.publications.through.objects.count(), 3)
        time.sleep(1)

        article.publications.add(p2, p1)
        state_time4 = article.publications.last_update_time()
        self.assertItemsEqual(article.publications.all(), [p1, p2, p3])
        self.assertEqual(article.publications.count(), 3)
        self.assertEqual(article.publications.through.objects.count(), 5)
        time.sleep(1)

        article.publications.remove(p2, p1)
        state_time5 = article.publications.last_update_time()
        self.assertItemsEqual(article.publications.all(), [p3])
        self.assertEqual(article.publications.count(), 1)
        self.assertEqual(article.publications.through.objects.count(), 5)
        time.sleep(1)

        article.publications = [p1, p2]
        state_time6 = article.publications.last_update_time()
        self.assertItemsEqual(article.publications.all(), [p1, p2])
        self.assertEqual(article.publications.count(), 2)
        self.assertEqual(article.publications.through.objects.count(), 7)
        time.sleep(1)

        article.publications.clear()
        state_time7 = article.publications.last_update_time()
        self.assertItemsEqual(article.publications.all(), [])
        self.assertEqual(article.publications.count(), 0)
        self.assertEqual(article.publications.through.objects.count(), 7)
        time.sleep(1)

        # test of history
        self.assertItemsEqual(article.publications.were_at(state_time1), [])
        self.assertItemsEqual(article.publications.were_at(state_time2), [p1, p2])
        self.assertItemsEqual(article.publications.were_at(state_time3), [p3])
        self.assertItemsEqual(article.publications.were_at(state_time4), [p1, p2, p3])
        self.assertItemsEqual(article.publications.were_at(state_time5), [p3])
        self.assertItemsEqual(article.publications.were_at(state_time6), [p1, p2])
        self.assertItemsEqual(article.publications.were_at(state_time7), [])

        # test of added_at and removed
        self.assertItemsEqual(article.publications.added_at(state_time2), [p1, p2])
        self.assertItemsEqual(article.publications.removed_at(state_time2), [])
        self.assertItemsEqual(article.publications.added_at(state_time3), [p3])
        self.assertItemsEqual(article.publications.removed_at(state_time3), [p1, p2])
        self.assertItemsEqual(article.publications.added_at(state_time4), [p1, p2])
        self.assertItemsEqual(article.publications.removed_at(state_time4), [])
        self.assertItemsEqual(article.publications.added_at(state_time5), [])
        self.assertItemsEqual(article.publications.removed_at(state_time5), [p1, p2])
        self.assertItemsEqual(article.publications.added_at(state_time6), [p1, p2])
        self.assertItemsEqual(article.publications.removed_at(state_time6), [p3])
        self.assertItemsEqual(article.publications.added_at(state_time7), [])
        self.assertItemsEqual(article.publications.removed_at(state_time7), [p1, p2])

        # test different arguments
        self.assertItemsEqual(article.publications.were_at(state_time4, only_pk=True), map(lambda o: o.pk, article.publications.were_at(state_time4)))
        with self.assertRaises(ValueError):
            article.publications.were_at(state_time5, unique=False)

        # test versions
        self.assertEqual(ManyToManyHistoryVersion.objects.count(), 6)
        for i in range(2, 8):
            state_time = locals()['state_time%d' % i]
            version = article.publications.get_version(time=state_time)
            self.assertItemsEqual(version.items,    article.publications.were_at(state_time))
            self.assertItemsEqual(version.added,    article.publications.added_at(state_time))
            self.assertItemsEqual(version.removed,  article.publications.removed_at(state_time))
            self.assertEqual(version.count,         article.publications.were_at(state_time).count())
            self.assertEqual(version.added_count,   article.publications.added_at(state_time).count())
            self.assertEqual(version.removed_count, article.publications.removed_at(state_time).count())

        article.publications_no_cache = [p1, p2]
        self.assertEqual(ManyToManyHistoryVersion.objects.count(), 6)

    def test_m2m_default_features(self):
        '''
        Build-in test from https://docs.djangoproject.com/en/dev/topics/db/examples/many_to_many/
        '''
        p1 = Publication(title='The Python Journal')
        p1.save()
        p2 = Publication(title='Science News')
        p2.save()
        p3 = Publication(title='Science Weekly')
        p3.save()

        a1 = Article(headline='Django lets you build Web apps easily')
        a1.save()
        a1.publications.add(p1)

        a2 = Article(headline='NASA uses Python')
        a2.save()
        a2.publications.add(p1, p2)
        a2.publications.add(p3)
        a2.publications.add(p3)
        with self.assertRaises(TypeError):
            a2.publications.add(a1)

        p4 = a2.publications.create(title='Highlights for Children')
        self.assertItemsEqual(a1.publications.all(), [p1])
        self.assertItemsEqual(a2.publications.all(), [p4, p2, p3, p1])

        self.assertItemsEqual(p2.article_set.all(), [a2])
        self.assertItemsEqual(p1.article_set.all(), [a1, a2])
        self.assertItemsEqual(Publication.objects.get(id=4).article_set.all(), [a2])

        self.assertItemsEqual(Article.objects.filter(publications__id=1), [a1, a2])
        self.assertItemsEqual(Article.objects.filter(publications__pk=1), [a1, a2])
        self.assertItemsEqual(Article.objects.filter(publications=1), [a1, a2])
        self.assertItemsEqual(Article.objects.filter(publications=p1), [a1, a2])

        self.assertItemsEqual(Article.objects.filter(publications__title__startswith="Science"), [a2, a2])
        self.assertItemsEqual(Article.objects.filter(publications__title__startswith="Science").distinct(), [a2])
        self.assertEqual(Article.objects.filter(publications__title__startswith="Science").count(), 2)
        self.assertEqual(Article.objects.filter(publications__title__startswith="Science").distinct().count(), 1)

        self.assertItemsEqual(Article.objects.filter(publications__in=[1,2]).distinct(), [a1, a2])
        self.assertItemsEqual(Article.objects.filter(publications__in=[p1,p2]).distinct(), [a1, a2])

        self.assertItemsEqual(Publication.objects.filter(id=1), [p1])
        self.assertItemsEqual(Publication.objects.filter(pk=1), [p1])
        self.assertItemsEqual(Publication.objects.filter(article__headline__startswith="NASA"), [p4, p3, p2, p1])

        self.assertItemsEqual(Publication.objects.filter(article__id=1), [p1])
        self.assertItemsEqual(Publication.objects.filter(article__pk=1), [p1])
        self.assertItemsEqual(Publication.objects.filter(article=1), [p1])
        self.assertItemsEqual(Publication.objects.filter(article=a1), [p1])

        self.assertItemsEqual(Publication.objects.filter(article__in=[1,2]).distinct(), [p1, p2, p3, p4])
        self.assertItemsEqual(Publication.objects.filter(article__in=[a1,a2]).distinct(), [p1, p2, p3, p4])
        self.assertItemsEqual(Article.objects.exclude(publications=p2), [a1])

        p1.delete()
        self.assertItemsEqual(Publication.objects.all(), [p2, p3, p4])
        a1 = Article.objects.get(pk=1)
        self.assertItemsEqual(a1.publications.all(), [])

        a2.delete()
        self.assertItemsEqual(Article.objects.all(), [a1])
        self.assertItemsEqual(p2.article_set.all(), [])

        a4 = Article(headline='NASA finds intelligent life on Earth')
        a4.save()
        p2.article_set.add(a4)
        p2.article_set.all()
        self.assertItemsEqual(p2.article_set.all(), [a4])
        self.assertItemsEqual(a4.publications.all(), [p2])

        new_article = p2.article_set.create(headline='Oxygen-free diet works wonders')
        self.assertItemsEqual(p2.article_set.all(), [a4, new_article])
        a5 = p2.article_set.all()[1]
        self.assertItemsEqual(a5.publications.all(), [p2])

        # Removing Publication from an Article:
        self.assertItemsEqual(p2.article_set.all(), [a4, a5])
        a4.publications.remove(p2)
        self.assertItemsEqual(p2.article_set.all(), [a5])
        self.assertItemsEqual(a4.publications.all(), [])

        # And from the other end:
        p2.article_set.remove(a5)
        self.assertItemsEqual(p2.article_set.all(), [])
        self.assertItemsEqual(a5.publications.all(), [])

        # Relation sets can be assigned. Assignment clears any existing set members:
        self.assertItemsEqual(a4.publications.all(), [])
        a4.publications = [p3]
        self.assertItemsEqual(a4.publications.all(), [p3])

        # Relation sets can be cleared:
        p2.article_set.clear()
        self.assertItemsEqual(p2.article_set.all(), [])

        # And you can clear from the other end:
        p2.article_set.add(a4, a5)
        self.assertItemsEqual(p2.article_set.all(), [a4, a5])
        self.assertItemsEqual(a4.publications.all(), [p2, p3])
        a4.publications.clear()
        self.assertItemsEqual(a4.publications.all(), [])
        self.assertItemsEqual(p2.article_set.all(), [a5])

        # Recreate the Article and Publication we have deleted:
        p1 = Publication(title='The Python Journal')
        p1.save()
        a2 = Article(headline='NASA uses Python')
        a2.save()
        a2.publications.add(p1, p2, p3)

        # Bulk delete some Publications - references to deleted publications should go:
        Publication.objects.filter(title__startswith='Science').delete()
        self.assertItemsEqual(Publication.objects.all(), [p4, p1])
        self.assertItemsEqual(Article.objects.all(), [a1, a2, a4, a5])
        self.assertItemsEqual(a2.publications.all(), [p1])

        # Bulk delete some articles - references to deleted objects should go:
        q = Article.objects.filter(headline__startswith='Django')
        self.assertItemsEqual(q, [a1])
        q.delete()

        # After the delete(), the QuerySet cache needs to be cleared, and the referenced objects should be gone:
        self.assertItemsEqual(q, [])
        self.assertItemsEqual(p1.article_set.all(), [a2])

        # An alternate to calling clear() is to assign the empty set:
        p1.article_set = []
        self.assertItemsEqual(p1.article_set.all(), [])

        a2.publications = [p1, p4]
        self.assertItemsEqual(a2.publications.all(), [p1, p4])
        a2.publications = []
        self.assertItemsEqual(a2.publications.all(), [])


# -*- coding: utf-8 -*-
from django.db import models
from descriptors import ManyRelatedObjectsHistoryDescriptor, ReverseManyRelatedObjectsHistoryDescriptor

__all__ = ['ManyToManyHistoryField']


class ManyToManyHistoryField(models.ManyToManyField):

    def __init__(self, *args, **kwargs):
        self.cache = kwargs.pop('cache', False)
        super(ManyToManyHistoryField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        '''
        Call super method and remove unique_together, add time fields and change descriptor class
        '''
        super(ManyToManyHistoryField, self).contribute_to_class(cls, name)

        self.rel.through._meta.unique_together = ()
        self.rel.through.add_to_class('time_from', models.DateTimeField(u'Datetime from', null=True, db_index=True))
        self.rel.through.add_to_class('time_to',  models.DateTimeField(u'Datetime to', null=True, db_index=True))
        # wrong behaviour of south
#        self.rel.through._meta.auto_created = False

        setattr(cls, self.name, ReverseManyRelatedObjectsHistoryDescriptor(self))

    def contribute_to_related_class(self, cls, related):
        '''
        Change descriptor class
        '''
        super(ManyToManyHistoryField, self).contribute_to_related_class(cls, related)

        # `swapped` attribute is not present before Django 1.5
        if not self.rel.is_hidden() and not getattr(related.model._meta, 'swapped', None):
            setattr(cls, related.get_accessor_name(), ManyRelatedObjectsHistoryDescriptor(related))

#     def _get_m2m_db_table(self, opts):
#         db_table = super(ManyToManyHistoryField, self)._get_m2m_db_table(opts)
#         return db_table + '_history'

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ["^m2m_history\.fields\.ManyToManyHistoryField"])
except ImportError:
    pass



# -*- coding: utf-8 -*-
from django.db import models
from django.db.models import Q
from django.contrib.contenttypes.models import ContentType
from django.db.models.fields.related import ManyRelatedObjectsDescriptor, ReverseManyRelatedObjectsDescriptor, cached_property, create_many_related_manager, router, signals
from signals import m2m_history_changed
from models import ManyToManyHistoryVersion
from datetime import datetime


def create_many_related_history_manager(superclass, rel):
    baseManagerClass = create_many_related_manager(superclass, rel)

    class ManyToManyHistoryThroughManager(baseManagerClass):

        # time of altering transaction
        time = None

        def get_version(self, time):
            return ManyToManyHistoryVersion.objects.get(content_type=ContentType.objects.get_for_model(self.instance), field_name=self.prefetch_cache_name, time=time)

        def get_time(self):
            if not self.time:
                self.time = datetime.now()
            return self.time

        def last_update_time(self):
            # TODO: optimize this method to one query
            qs = self.get_query_set_through()
            try:
                time_to = qs.exclude(time_to=None).order_by('-time_to')[0].time_to
                time_from = qs.exclude(time_from=None).order_by('-time_from')[0].time_from
                return time_to if time_to > time_from else time_from
            except IndexError:
                return qs.exclude(time_from=None).order_by('-time_from')[0].time_from

        def get_user_ids_of_period(self, group, date_from, date_to, field=None, unique=True):

            if field is None:
                # TODO: make normal filtering
                kwargs = {'time_entered': None, 'time_left': None} \
                    | {'time_entered__lte': date_from, 'time_left': None} \
                    | {'time_entered': None, 'time_left__gte': date_to}
            elif field in ['left','entered']:
                kwargs = {'time_%s__gt' % field: date_from, 'time_%s__lte' % field: date_to}
            else:
                raise ValueError("Attribute `field` should be equal to 'left' of 'entered'")

            qs = self.filter(group=group, **kwargs)
            return self._prepare_qs(qs, unique)

        def _prepare_queryset(self, queryset, only_pk=False, unique=True):
            queryset = queryset.values_list(self.target_field_name, flat=True)
            if not only_pk:
                if unique == False:
                    raise ValueError("Argument `unique` should be True if argument only_pk is False")
                db = router.db_for_write(self.through, instance=self.instance)
                queryset = super(ManyToManyHistoryThroughManager, self).get_query_set().using(db).filter(pk__in=queryset)

            if unique:
                queryset = queryset.distinct()
            return queryset

        def get_query_set(self, **kwargs):
            queryset = self.get_query_set_through().filter(time_to=None)
            return self._prepare_queryset(queryset, **kwargs)

        def get_query_set_through(self):
            db = router.db_for_write(self.through, instance=self.instance)
            qs = self.through._default_manager.using(db).filter(**{
                self.source_field_name: self._fk_val,
            })
            return qs

        def were_at(self, time, **kwargs):
            queryset = self.get_query_set_through()
            queryset = queryset.filter(
                Q(time_from=None,        time_to=None) | \
                Q(time_from=None,        time_to__gt=time) | \
                Q(time_from__lte=time,   time_to=None) | \
                Q(time_from__lte=time,   time_to__gt=time))
            return self._prepare_queryset(queryset, **kwargs)

        def added_at(self, time, **kwargs):
            queryset = self.get_query_set_through().filter(time_from=time)
            return self._prepare_queryset(queryset, **kwargs)

        def removed_at(self, time, **kwargs):
            queryset = self.get_query_set_through().filter(time_to=time)
            return self._prepare_queryset(queryset, **kwargs)

        def clear(self, *objs):
            self._clear_items(self.source_field_name, self.target_field_name, *objs)

            # If this is a symmetrical m2m relation to self, clear the mirror entry in the m2m table
            if self.symmetrical:
                self._clear_items(self.target_field_name, self.source_field_name, *objs)
        clear.alters_data = True

        def send_signal(self, source_field_name, action, ids):
            if self.reverse or source_field_name == self.source_field_name:
                # Don't send the signal when we are inserting the
                # duplicate data row for symmetrical reverse entries.
                db = router.db_for_write(self.through, instance=self.instance)

                signals.m2m_changed.send(sender=self.through, action=action,
                    instance=self.instance, reverse=self.reverse,
                    model=self.model, pk_set=ids, using=db)

                m2m_history_changed.send(sender=self.through, action=action,
                    instance=self.instance, reverse=self.reverse,
                    model=self.model, pk_set=ids, using=db,
                    field_name=self.prefetch_cache_name, time=self.get_time())

        def _add_items(self, source_field_name, target_field_name, *objs):
            # source_field_name: the PK fieldname in join table for the source object
            # target_field_name: the PK fieldname in join table for the target object
            # *objs - objects to add. Either object instances, or primary keys of object instances.

            # If there aren't any objects, there is nothing to do.
            from django.db.models import Model
            if objs:
                new_ids = set()
                for obj in objs:
                    if isinstance(obj, self.model):
                        if not router.allow_relation(obj, self.instance):
                            raise ValueError('Cannot add "%r": instance is on database "%s", value is on database "%s"' %
                                               (obj, self.instance._state.db, obj._state.db))
                        fk_val = self._get_fk_val(obj, target_field_name)
                        if fk_val is None:
                            raise ValueError('Cannot add "%r": the value for field "%s" is None' %
                                             (obj, target_field_name))
                        new_ids.add(self._get_fk_val(obj, target_field_name))
                    elif isinstance(obj, Model):
                        raise TypeError("'%s' instance expected, got %r" % (self.model._meta.object_name, obj))
                    else:
                        new_ids.add(obj)
                db = router.db_for_write(self.through, instance=self.instance)
                vals = self.through._default_manager.using(db).values_list(target_field_name, flat=True)
                vals = vals.filter(**{
                    source_field_name: self._fk_val,
                    '%s__in' % target_field_name: new_ids,
                    'time_to': None,
                })
                new_ids = new_ids - set(vals)

                self.send_signal(source_field_name, 'pre_add', new_ids)

                # Add the ones that aren't there already
                self.through._default_manager.using(db).bulk_create([
                    self.through(**{
                        '%s_id' % source_field_name: self._fk_val,
                        '%s_id' % target_field_name: obj_id,
                        'time_from': self.get_time(),
                    })
                    for obj_id in new_ids
                ])

                self.send_signal(source_field_name, 'post_add', new_ids)

        def _remove_items(self, source_field_name, target_field_name, *objs):
            # source_field_name: the PK colname in join table for the source object
            # target_field_name: the PK colname in join table for the target object
            # *objs - objects to remove

            # If there aren't any objects, there is nothing to do.
            if objs:
                # Check that all the objects are of the right type
                old_ids = set()
                for obj in objs:
                    if isinstance(obj, self.model):
                        old_ids.add(self._get_fk_val(obj, target_field_name))
                    else:
                        old_ids.add(obj)
                # Work out what DB we're operating on
                db = router.db_for_write(self.through, instance=self.instance)

                self.send_signal(source_field_name, 'pre_remove', old_ids)

                # Remove the specified objects from the join table
                qs = self.through._default_manager.using(db).filter(**{
                    source_field_name: self._fk_val,
                    '%s__in' % target_field_name: old_ids,
                    'time_to': None,
                })
                qs.update(time_to=self.get_time())

                self.send_signal(source_field_name, 'post_remove', old_ids)

        def _clear_items(self, source_field_name, target_field_name, *objs):
            db = router.db_for_write(self.through, instance=self.instance)
            # source_field_name: the PK colname in join table for the source object

            self.send_signal(source_field_name, 'pre_clear', None)

            qs = self.through._default_manager.using(db).filter(**{
                source_field_name: self._fk_val,
                'time_to': None,
            }).exclude(**{
                '%s__in' % target_field_name: [obj.pk for obj in objs]
            })
            qs.update(time_to=self.get_time())

            self.send_signal(source_field_name, 'post_clear', set(self.removed_at(self.get_time(), only_pk=True)))

    return ManyToManyHistoryThroughManager


class ReverseManyRelatedObjectsHistoryDescriptor(ReverseManyRelatedObjectsDescriptor):
    @cached_property
    def related_manager_cls(self):
        '''
        Difference from super method is return our own manager inherited from the build-in
        '''
        return create_many_related_history_manager(
            self.field.rel.to._default_manager.__class__,
            self.field.rel
        )

    def __set__(self, instance, value):
        '''
        Difference from super method is send value to `clear` method as well as to `add` method
        '''
        if instance is None:
            raise AttributeError("Manager must be accessed via instance")

        if not self.field.rel.through._meta.auto_created:
            opts = self.field.rel.through._meta
            raise AttributeError("Cannot set values on a ManyToManyField which specifies an intermediary model.  Use %s.%s's Manager instead." % (opts.app_label, opts.object_name))

        manager = self.__get__(instance)
        manager.clear(*value)
        manager.add(*value)


class ManyRelatedObjectsHistoryDescriptor(ManyRelatedObjectsDescriptor):
    @cached_property
    def related_manager_cls(self):
        '''
        Difference from super method is return our own manager inherited from the build-in
        '''
        return create_many_related_history_manager(
            self.related.model._default_manager.__class__,
            self.related.field.rel
        )

    def __set__(self, instance, value):
        '''
        Difference from super method is send value to `clear` method as well as to `add` method
        '''
        if instance is None:
            raise AttributeError("Manager must be accessed via instance")

        if not self.related.field.rel.through._meta.auto_created:
            opts = self.related.field.rel.through._meta
            raise AttributeError("Cannot set values on a ManyToManyField which specifies an intermediary model. Use %s.%s's Manager instead." % (opts.app_label, opts.object_name))

        manager = self.__get__(instance)
        manager.clear(*value)
        manager.add(*value)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'ManyToManyHistoryVersion'
        db.create_table(u'm2m_history_manytomanyhistoryversion', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='m2m_history_versions', to=orm['contenttypes.ContentType'])),
            ('object_id', self.gf('django.db.models.fields.BigIntegerField')(db_index=True)),
            ('field_name', self.gf('django.db.models.fields.CharField')(max_length=50, db_index=True)),
            ('time', self.gf('django.db.models.fields.DateTimeField')(db_index=True)),
            ('count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('added_count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('removed_count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
        ))
        db.send_create_signal(u'm2m_history', ['ManyToManyHistoryVersion'])

        # Adding unique constraint on 'ManyToManyHistoryVersion', fields ['content_type', 'object_id', 'field_name', 'time']
        db.create_unique(u'm2m_history_manytomanyhistoryversion', ['content_type_id', 'object_id', 'field_name', 'time'])


    def backwards(self, orm):
        # Removing unique constraint on 'ManyToManyHistoryVersion', fields ['content_type', 'object_id', 'field_name', 'time']
        db.delete_unique(u'm2m_history_manytomanyhistoryversion', ['content_type_id', 'object_id', 'field_name', 'time'])

        # Deleting model 'ManyToManyHistoryVersion'
        db.delete_table(u'm2m_history_manytomanyhistoryversion')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'm2m_history.manytomanyhistoryversion': {
            'Meta': {'unique_together': "(('content_type', 'object_id', 'field_name', 'time'),)", 'object_name': 'ManyToManyHistoryVersion'},
            'added_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'m2m_history_versions'", 'to': u"orm['contenttypes.ContentType']"}),
            'count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'field_name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'db_index': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_id': ('django.db.models.fields.BigIntegerField', [], {'db_index': 'True'}),
            'removed_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'time': ('django.db.models.fields.DateTimeField', [], {'db_index': 'True'})
        }
    }

    complete_apps = ['m2m_history']